'''
Migration support as features in MongoWorld 2016
From the Polls to the Trolls.

The Manager class provides a general purpose support
for migrating documents to a target version through
a series of migration functions.
'''

import re
import itertools

from six.moves import range

from more_itertools import recipes


class Manager(object):
	"""
	A manager for facilitating the registration of migration functions
	and applying those migrations to documents.

	To use, implement migration functions to and from each adjacent
	version of your schema and decorate each with the register
	classmethod. For example:

	>>> @Manager.register
	... def v1_to_2(manager, doc):
	...     doc['foo'] = 'bar'
	>>> @Manager.register
	... def v2_to_3(manager, doc):
	...     doc['foo'] = doc['foo'] + ' baz'

	Note that in addition to the document, the migration manager is also
	passed to the migration function, allowing for other context to be
	made available during the migration.

	To create a manager for migrating documents to version 3:

	>>> mgr = Manager(3)

	Then, use the manager to migrate a document to a target version.

	>>> v1_doc = dict(version=1, data='stub')
	>>> v3_doc = mgr.migrate_doc(v1_doc)

	>>> v3_doc['version']
	3
	>>> v3_doc['foo']
	'bar baz'

	Note that the document is modified in place:

	>>> v1_doc is v3_doc
	True

	>>> Manager._upgrade_funcs.clear()
	"""

	version_attribute_name = 'version'
	_upgrade_funcs = set()

	def __init__(self, target_version):
		self.target_version = target_version

	@classmethod
	def register(cls, func):
		"""
		Decorate a migration function with this method
		to make it available for migrating cases.
		"""
		cls._add_version_info(func)
		cls._upgrade_funcs.add(func)
		return func

	@staticmethod
	def _add_version_info(func):
		"""
		Add .source and .target attributes to the registered function.
		"""
		pattern = r'v(?P<source>\d+)_to_(?P<target>\d+)$'
		match = re.match(pattern, func.__name__)
		if not match:
			raise ValueError("migration function name must match " + pattern)
		func.source, func.target = map(int, match.groups())

	def migrate_doc(self, doc):
		"""
		Migrate the doc from its current version to the target version
		and return it.
		"""
		orig_ver = doc.get(self.version_attribute_name, 0)
		funcs = self._get_migrate_funcs(orig_ver, self.target_version)
		for func in funcs:
			func(self, doc)
			doc[self.version_attribute_name] = func.target
		return doc

	@classmethod
	def _get_migrate_funcs(cls, orig_version, target_version):
		"""
		>>> @Manager.register
		... def v1_to_2(manager, doc):
		...     doc['foo'] = 'bar'
		>>> @Manager.register
		... def v2_to_1(manager, doc):
		...     del doc['foo']
		>>> @Manager.register
		... def v2_to_3(manager, doc):
		...     doc['foo'] = doc['foo'] + ' baz'
		>>> funcs = list(Manager._get_migrate_funcs(1, 3))
		>>> len(funcs)
		2
		>>> funcs == [v1_to_2, v2_to_3]
		True
		>>> funcs = list(Manager._get_migrate_funcs(2, 1))
		>>> len(funcs)
		1
		>>> funcs == [v2_to_1]
		True

		>>> Manager._upgrade_funcs.clear()
		"""
		direction = 1 if target_version > orig_version else -1
		versions = range(orig_version, target_version + direction, direction)
		transitions = recipes.pairwise(versions)
		return itertools.starmap(cls._get_func, transitions)

	@classmethod
	def _get_func(cls, source_ver, target_ver):
		"""
		Return exactly one function to convert from source to target
		"""
		matches = (
			func for func in cls._upgrade_funcs
			if func.source == source_ver and func.target == target_ver
		)
		try:
			match, = matches
		except ValueError:
			msg = "No migration from {source_ver} to {target_ver}"
			raise ValueError(msg.format(**locals()))
		return match



"""
Script to move a subset of GridFS files from one db
(or collection) to another.
"""

import sys
import logging
import argparse
import itertools

from six.moves import map

from jaraco.ui import progress
from more_itertools.recipes import consume

from jaraco.mongodb import helper


log = logging.getLogger()


def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('source_gfs', type=helper.connect_gridfs)
	parser.add_argument('dest_gfs', type=helper.connect_gridfs)
	parser.add_argument('--include',
		help="a filter of files (regex) to include",
	)
	parser.add_argument('--delete', default=False, action="store_true",
		help="delete files after moving",
	)
	parser.add_argument('--limit', type=int)
	return parser.parse_args()


class FileMove:
	include = None
	delete = False
	limit = None

	def __init__(self, **params):
		vars(self).update(**params)

	@property
	def filter(self):
		if not self.include:
			return
		return dict(filename={"$regex": self.include})

	@property
	def source_coll(self):
		return self.source_gfs._GridFS__collection

	@property
	def dest_coll(self):
		return self.dest_gfs._GridFS__collection

	def run(self):
		files = self.source_coll.files.find(self.filter,
			batch_size=1)
		limit_files = itertools.islice(files, self.limit)
		count = min(files.count(), self.limit or float('inf'))
		bar = progress.TargetProgressBar(count)
		to_process = map(self.process, bar.iterate(limit_files))
		consume(to_process)

	def process(self, file):
		chunks = self.source_coll.chunks.find(dict(files_id=file['_id']))
		for chunk in chunks:
			self.dest_coll.chunks.insert(chunk)
		self.dest_coll.files.insert(file)
		self.delete and self.source_gfs.delete(file['_id'])


def run():
	logging.basicConfig(stream=sys.stderr, level=logging.INFO)
	args = get_args()

	mover = FileMove(**vars(args))
	mover.run()


if __name__ == '__main__':
	run()



import shlex

import pmxbot

from . import sharding


@pmxbot.core.command("create-db-in-shard")
def cdbs(client, event, channel, nick, rest):
        """
        Create a database in a shard. !create-db-in-shard {db name} {shard name}
        """
        db_name, shard = shlex.split(rest)
        return sharding.create_db_in_shard(db_name, shard)



from __future__ import unicode_literals, absolute_import

import argparse
import time
import json
import logging
import pymongo
import bson
import re
import textwrap
import collections

import jaraco.logging
from pymongo.cursor import CursorType
from jaraco.itertools import always_iterable


class Extend(argparse.Action):
    """
    Argparse action to take an nargs=* argument
    and add any values to the existing value.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        getattr(namespace, self.dest).extend(values)


def parse_args(*args, **kwargs):
    """
    Parse the args for the command.

    It should be possible for one to specify '--ns', '-x', and '--rename'
    multiple times:

    >>> args = parse_args(['--ns', 'foo', 'bar', '--ns', 'baz'])
    >>> args.ns
    ['foo', 'bar', 'baz']

    >>> parse_args(['-x', '--exclude']).exclude
    []

    >>> renames = parse_args(['--rename', 'a=b', '--rename', 'b=c']).rename
    >>> len(renames)
    2
    >>> type(renames)
    <class 'jaraco.mongodb.oplog.Renamer'>
    """
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument("--help",
        help="show usage information",
        action="help")

    parser.add_argument("--source", metavar="host[:port]",
        help="host to pull from")

    parser.add_argument('--oplogns', default='local.oplog.rs',
        help="source namespace for oplog")

    parser.add_argument("--dest", metavar="host[:port]",
        default="localhost",
        help="host to push to (<set name>/s1,s2 for sets)")

    parser.add_argument("-s", "--seconds", type=int, default=None,
        help="""seconds to go back. If not set, try read
        timestamp from --resume-file. If the file not found,
        assume --seconds=86400 (24 hours)""")

    parser.add_argument("-f", "--follow", action="store_true",
        help="wait for new data in oplog, run forever.")

    parser.add_argument("--ns", nargs="*", default=[],
        action=Extend,
        help="this namespace(s) only ('dbname' or 'dbname.coll')")

    parser.add_argument("-x", "--exclude", nargs="*", default=[],
        action=Extend,
        help="exclude namespaces ('dbname' or 'dbname.coll')")

    parser.add_argument("--rename", nargs="*", default=[],
        metavar="ns_old=ns_new",
        type=Renamer.item,
        action=Extend,
        help="rename namespaces before processing on dest")

    parser.add_argument("--dry-run", default=False,
        action="store_true",
        help="suppress application of ops")

    help = textwrap.dedent("""
        resume from timestamp read from this file and
        write last processed timestamp back to this file
        (default is %(default)s).
        Pass empty string or 'none' to disable this
        feature.
        """)
    parser.add_argument("--resume-file", default="mongooplog.ts",
        metavar="FILENAME", type=string_none,
        help=help,
    )
    jaraco.logging.add_arguments(parser)

    args = parser.parse_args(*args, **kwargs)
    args.rename = Renamer(args.rename)
    return args


class Renamer(dict):
    """
    >>> specs = [
    ...      'a=b',
    ...      'alpha=gamma',
    ...  ]
    >>> renames = Renamer(map(Renamer.item, specs))
    >>> op = dict(ns='a.a')
    >>> renames(op)
    >>> op['ns']
    'b.a'
    >>> renames.affects('alpha.foo')
    True
    >>> renames.affects('b.gamma')
    False
    """

    def invoke(self, op):
        """
        Replace namespaces in op based on keys/values in self.
        """
        for old_ns, new_ns in self.items():
            if old_ns.match(op['ns']):
                ns = old_ns.sub(new_ns, op['ns']).rstrip(".")
                logging.debug("renaming %s to %s", op['ns'], ns)
                op['ns'] = ns
            if op['ns'].endswith('.system.indexes'):
                # index operation; update ns in the op also.
                self.invoke(op['o'])
    __call__ = invoke

    @classmethod
    def from_specs(cls, specs):
        return cls(map(cls.item, always_iterable(specs)))

    @staticmethod
    def item(spec):
        """
        Return a pair of old namespace (regex) to the new namespace (string).

        spec should be a pair separated by equal sign ('=').
        """
        old_ns, new_ns = spec.split('=')
        regex = re.compile(r"^{0}(\.|$)".format(re.escape(old_ns)))

        return regex, new_ns + "."

    def affects(self, ns):
        """
        Return True if this renamer affects the indicated namespace.
        """
        return any(exp.match(ns) for exp in self)


def string_none(value):
    """
    Convert the string 'none' to None
    """
    is_string_none = not value or value.lower() == 'none'
    return None if is_string_none else value


def _calculate_start(args):
    """
    Return the start time as a bson timestamp.
    """
    utcnow = int(time.time())

    if args.seconds:
        return bson.timestamp.Timestamp(utcnow - args.seconds, 0)

    day_ago = bson.timestamp.Timestamp(utcnow - 24*60*60, 0)
    saved_ts = read_ts(args.resume_file)
    spec_ts = increment_ts(saved_ts) if saved_ts else None
    return spec_ts or day_ago


def _same_instance(client1, client2):
    """
    Return True if client1 and client2 appear to reference the same
    MongoDB instance.
    """
    return client1._topology_settings.seeds == client2._topology_settings.seeds


def _full_rename(args):
    """
    Return True only if the arguments passed specify exact namespaces
    and to conduct a rename of every namespace.
    """
    return (
        args.ns and
        all(map(args.rename.affects, args.ns))
    )


def _resolve_shard(client):
    """
    The destination cannot be a mongoS instance, as applyOps is
    not an allowable command for mongoS instances, so if the
    client is a connection to a mongoS instance, raise an error
    or resolve the replica set.
    """
    status = client.admin.command('serverStatus')
    if status['process'] == 'mongos':
        raise RuntimeError("Destination cannot be mongos")
    return client


def main():
    args = parse_args()
    log_format = '%(asctime)s - %(levelname)s - %(message)s'
    jaraco.logging.setup(args, format=log_format)

    logging.info("going to connect")

    src = pymongo.MongoClient(args.source)
    dest = _resolve_shard(pymongo.MongoClient(args.dest))

    if _same_instance(src, dest) and not _full_rename(args):
        logging.error(
            "source and destination hosts can be the same only "
            "when both --ns and --rename arguments are given")
        raise SystemExit(1)

    logging.info("connected")

    start = _calculate_start(args)

    logging.info("starting from %s (%s)", start, start.as_datetime())
    db_name, sep, coll_name = args.oplogns.partition('.')
    oplog_coll = src[db_name][coll_name]
    num = 0

    class_ = TailingOplog if args.follow else Oplog
    generator = class_(oplog_coll)

    try:
        for num, doc in enumerate(generator.since(start)):
            _handle(dest, doc, args, num)
            last_handled = doc
        logging.info("all done")
    except KeyboardInterrupt:
        logging.info("Got Ctrl+C, exiting...")
    finally:
        if 'last_handled' in locals():
            save_ts(last_handled['ts'], args.resume_file)

def _handle(dest, op, args, num):
    # Skip "no operation" items
    if op['op'] == 'n':
        return

    # Update status
    ts = op['ts']
    if not num % 1000:
        save_ts(ts, args.resume_file)
        logging.info(
            "%s\t%s\t%s -> %s",
            num, ts.as_datetime(),
            op.get('op'),
            op.get('ns'),
        )

    # Skip excluded namespaces or namespaces that does not match --ns
    excluded = any(op['ns'].startswith(ns) for ns in args.exclude)
    included = any(op['ns'].startswith(ns) for ns in args.ns)

    if excluded or (args.ns and not included):
        logging.log(logging.DEBUG-1, "skipping ns %s", op['ns'])
        return

    args.rename(op)

    logging.debug("applying op %s", op)
    try:
        apply(dest, op) if not args.dry_run else None
    except pymongo.errors.OperationFailure as e:
        msg = '{e!r} applying {op}'.format(**locals())
        logging.warning(msg)


def apply(db, op):
    """
    Apply operation in db
    """
    dbname = op['ns'].split('.')[0] or "admin"
    db[dbname].command("applyOps", [op])


class Oplog(object):
    find_params = {}

    def __init__(self, coll):
        self.coll = coll.with_options(
            codec_options=bson.CodecOptions(
                document_class=collections.OrderedDict,
            ),
        )

    def get_latest_ts(self):
        cur = self.coll.find().sort('$natural', pymongo.DESCENDING).limit(-1)
        latest_doc = next(cur)
        return latest_doc['ts']

    def query(self, spec):
        return self.coll.find(spec, **self.find_params)

    def since(self, ts):
        """
        Query the oplog for items since ts and then return
        """
        spec = {'ts': {'$gte': ts}}
        cursor = self.query(spec)
        while True:
            # todo: trap InvalidDocument errors:
            # except bson.errors.InvalidDocument as e:
            #  logging.info(repr(e))
            for doc in cursor:
                yield doc
            if not cursor.alive:
                break
            time.sleep(1)


class TailingOplog(Oplog):
    find_params = dict(
        cursor_type=CursorType.TAILABLE_AWAIT,
        oplog_replay=True,
    )

    def since(self, ts):
        """
        Tail the oplog, starting from ts.
        """
        while True:
            items = super(TailingOplog, self).since(ts)
            for doc in items:
                yield doc
                ts = doc['ts']


def save_ts(ts, filename):
    """Save last processed timestamp to file. """
    try:
        if filename:
            with open(filename, 'w') as f:
                obj = {"ts": {"time": ts.time, "inc":  ts.inc}}
                json.dump(obj, f)
    except IOError:
        pass


def read_ts(filename):
    """Read last processed timestamp from file. Return next timestamp that
    need to be processed, that is timestamp right after last processed one.
    """
    try:
        with open(filename, 'r') as f:
            data = json.load(f)['ts']
        return bson.Timestamp(data['time'], data['inc'])
    except (IOError, KeyError):
        pass


def increment_ts(ts):
    """
    Return a new ts with an incremented .inc.
    """
    return bson.Timestamp(ts.time, ts.inc + 1)


if __name__ == '__main__':
    main()



"""
Script to repair broken GridFS files. It handles

 - Removing files with missing chunks.
"""

import sys
import logging
import argparse

from six.moves import filter, map

import gridfs
from jaraco.ui import progress
from more_itertools.recipes import consume
from jaraco.itertools import Counter

from jaraco.mongodb import helper
from jaraco.context import ExceptionTrap


log = logging.getLogger()


def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('db', type=helper.connect_gridfs)
	return parser.parse_args()


class FileRepair:
	def __init__(self, gfs):
		self.gfs = gfs
		db = gfs._GridFS__database
		coll = gfs._GridFS__collection
		bu_coll_name = coll.name + '-saved'
		self.backup_coll = db[bu_coll_name]

	def run(self):
		files = self.gfs.list()
		bar = progress.TargetProgressBar(len(files))
		processed_files = map(self.process, bar.iterate(files))
		errors = filter(None, processed_files)
		counter = Counter(errors)
		consume(map(self.handle_trap, counter))
		return counter

	def process(self, filename):
		file = self.gfs.get_last_version(filename)
		with ExceptionTrap(gridfs.errors.CorruptGridFile) as trap:
			file.read(1)
		trap.filename = filename
		return trap

	def handle_trap(self, trap):
		cls, exc, tb = trap.exc_info
		spec = dict(filename=trap.filename)
		for file_doc in self.gfs._GridFS__files.find(spec):
			self.backup_coll.files.insert(file_doc)
			chunk_spec = dict(files_id=file_doc['_id'])
			for chunk in self.gfs._GridFS__chunks.find(chunk_spec):
				self.backup_coll.chunks.insert(chunk)
		log.info("Removing %s (%s)", trap.filename, exc)
		self.gfs.delete(spec)


def run():
	logging.basicConfig(stream=sys.stderr, level=logging.INFO)
	args = get_args()

	repair = FileRepair(args.db)
	counter = repair.run()

	log.info("Removed %s corrupt files.", counter.count)


if __name__ == '__main__':
	run()



import shlex

import pytest
try:
	import pymongo
except ImportError:
	pass

from . import service


def pytest_addoption(parser):
	parser.addoption('--mongod-args',
		help="Arbitrary arguments to mongod")


@pytest.yield_fixture(scope='session')
def mongodb_instance():
	if 'pymongo' not in globals():
		pytest.skip("pymongo not available")

	params_raw = pytest.config.getoption('mongod_args') or ''
	params = shlex.split(params_raw)
	service.MongoDBInstance.mongod_args += tuple(params)
	try:
		instance = service.MongoDBInstance()
		instance.log_root = ''
		instance.start()
		pymongo.MongoClient(instance.get_connect_hosts())
		yield instance
	except Exception as err:
		pytest.skip("MongoDB not available ({err})".format(**locals()))
	instance.stop()

@pytest.fixture(scope='session')
def mongodb_uri(mongodb_instance):
	return mongodb_instance.get_uri()



import socket
import operator

import pymongo
from jaraco.text import namespace_format as nf

hostname = socket.gethostname()
by_id = operator.itemgetter('_id')


def get_ids(collection):
        return map(by_id, collection.find(projection=['_id']))


def create_db_in_shard(db_name, shard, client=None):
        """
        In a sharded cluster, create a database in a particular shard.
        """
        client = client or pymongo.MongoClient()
        # flush the router config to ensure it's not stale
        res = client.admin.command('flushRouterConfig')
        if not res.get('ok'):
                raise RuntimeError("unable to flush router config")
        if shard not in get_ids(client.config.shards):
                raise ValueError(nf("Unknown shard {shard}"))
        if db_name in get_ids(client.config.databases):
                raise ValueError("database already exists")
        # MongoDB doesn't have a 'create database' command, so insert an
        #  item into a collection and then drop the collection.
        client[db_name].foo.insert({'foo': 1})
        client[db_name].foo.drop()
        if client[db_name].collection_names():
                raise ValueError("database has collections")
        res = client.admin.command('movePrimary', value=db_name, to=shard)
        if not res.get('ok'):
                raise RuntimeError(str(res))
        return nf("Successfully created {db_name} in {shard} via {hostname}")



import time
import re
import argparse

from jaraco.mongodb import helper

def is_index_op(op):
	return op.get('query', {}).get('createIndexes')

def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('db')
	return parser.parse_args()

def run():
	db = helper.connect_db(get_args().db)

	while True:
		ops = db.current_op()['inprog']
		index_op = next(filter(is_index_op, ops), None)
		if not index_op:
			print("No index operations in progress")
			break
		msg = index_op['msg']
		name = index_op['query']['indexes'][0]['name']
		pat = re.compile('Index Build( \(background\))?')
		msg = pat.sub(name, msg, count=1)
		print(msg, end='\r')
		time.sleep(5)

__name__ == '__main__' and run()



import datetime

class NeverExpires(object):
    def expired(self):
        return False

class Timer(object):
    """
    A simple timer that will indicate when an expiration time has passed.
    """
    def __init__(self, expiration):
        self.expiration = expiration

    @classmethod
    def after(cls, elapsed):
        """
        Return a timer that will expire after `elapsed` passes.
        """
        return cls(datetime.datetime.utcnow() + elapsed)

    def expired(self):
        return datetime.datetime.utcnow() >= self.expiration



import re


def all_databases(client, exclude=['local']):
	"""
	Yield all databases except excluded (default
	excludes 'local').
	"""
	return (
		client[db_name]
		for db_name in client.database_names()
		if db_name not in exclude
	)


def all_collections(db):
	"""
	Yield all non-sytem collections in db.
	"""
	include_pattern = r'(?!system\.)'
	return (
		db[name]
		for name in db.collection_names()
		if re.match(include_pattern, name)
	)


def purge_collection(coll):
	coll.remove(None)


def safe_purge_collection(coll):
	"""
	Cannot remove documents from capped collections
	in later versions of MongoDB, so drop the
	collection instead.
	"""
	op = (
		drop_collection
		if coll.options().get('capped', False)
		else purge_collection
	)
	return op(coll)


def drop_collection(coll):
	coll.database.drop_collection(coll.name)


def purge_all_databases(client, op=drop_collection):
	collections = (
		coll
		for db in all_databases(client)
		for coll in all_collections(db)
	)

	for coll in collections:
		op(coll)



from jaraco.mongodb import manage

def test_purge_all_databases(mongodb_instance):
	client = mongodb_instance.get_connection()
	client.test_db.test_coll.insert({'a': 1})
	client.test_db2.test_coll.insert({'b': 2})
	manage.purge_all_databases(client)
	indexes = {'system.indexes'}
	assert set(client.test_db.collection_names()) <= indexes
	assert set(client.test_db2.collection_names()) <= indexes



#!/usr/bin/env python

# Project skeleton maintained at https://github.com/jaraco/skeleton

import io
import sys

import setuptools

with io.open('README.rst', encoding='utf-8') as readme:
	long_description = readme.read()

needs_pytest = {'pytest', 'test'}.intersection(sys.argv)
pytest_runner = ['pytest_runner'] if needs_pytest else []
needs_sphinx = {'release', 'build_sphinx', 'upload_docs'}.intersection(sys.argv)
sphinx = ['sphinx', 'rst.linker'] if needs_sphinx else []
needs_wheel = {'release', 'bdist_wheel'}.intersection(sys.argv)
wheel = ['wheel'] if needs_wheel else []

name = 'jaraco.mongodb'
description = 'Routines and classes supporting MongoDB environments'

setup_params = dict(
	name=name,
	use_scm_version=True,
	author="Jason R. Coombs",
	author_email="jaraco@jaraco.com",
	description=description or name,
	long_description=long_description,
	url="https://github.com/jaraco/" + name,
	packages=setuptools.find_packages(),
	include_package_data=True,
	namespace_packages=name.split('.')[:-1],
	install_requires=[
		'pymongo',
		'six',
		'python-dateutil',
		'jaraco.services',
		'portend',
		'jaraco.itertools',
		'jaraco.functools',
		'jaraco.ui',
		'jaraco.context',
		'more_itertools',
		'jaraco.logging',
	],
	extras_require={
	},
	setup_requires=[
		'setuptools_scm>=1.9',
	] + pytest_runner + sphinx + wheel,
	tests_require=[
		'pytest>=2.8',
		'cherrypy',
	],
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3.3",
		"Programming Language :: Python :: 3.4",
		"Programming Language :: Python :: 3.5",
		"Framework :: Pytest",
	],
	entry_points={
		'pytest11': [
			'MongoDB = jaraco.mongodb.fixtures',
		],
		'pmxbot_handlers': [
			'create in MongoDB shard = jaraco.mongodb.pmxbot',
		],
	},
)
if __name__ == '__main__':
	setuptools.setup(**setup_params)



import bson
import pytest
import jaraco.itertools

from jaraco.mongodb import oplog
from jaraco.mongodb import service


class TestReplacer:
	def test_rename_index_op_ns(self):
		"""
		As an index operation references namespaces,
		when performing a rename operation, it's also important
		to rename the ns in the op itself.
		"""
		op = {
			'ts': bson.Timestamp(1446495808, 3),
			'ns': 'airportlocker.system.indexes',
			'op': 'i',
			'o': {
				'ns': 'airportlocker.luggage.chunks',
				'key': {'files_id': 1, 'n': 1},
				'name': 'files_id_1_n_1', 'unique': True
			},
		}

		ren = oplog.Renamer.from_specs("airportlocker=airportlocker-us")
		ren(op)

		assert op['ns'] == 'airportlocker-us.system.indexes'
		assert op['o']['ns'] == 'airportlocker-us.luggage.chunks'


def make_replicaset(request):
	try:
		r_set = service.MongoDBReplicaSet()
		r_set.log_root = ''
		r_set.start()
		r_set.get_connection()
		request.addfinalizer(r_set.stop)
	except Exception as err:
		pytest.skip("MongoDB not available ({err})".format(**locals()))
	return r_set


@pytest.fixture
def replicaset_factory(request):
	"""
	Return a factory that can generate MongoDB replica sets
	"""
	return jaraco.itertools.infiniteCall(make_replicaset, request)


class TestOplogReplication:
	def test_index_deletion(self, replicaset_factory):
		"""
		A delete index operation should be able to be applied to a replica
		"""
		source = next(replicaset_factory).get_connection()
		dest = next(replicaset_factory).get_connection()
		source.index_deletion_test.stuff.ensure_index("foo")
		dest.index_deletion_test.stuff.ensure_index("foo")
		source_oplog = oplog.Oplog(source.local.oplog.rs)
		begin_ts = source_oplog.get_latest_ts()
		source.index_deletion_test.stuff.drop_index("foo_1")
		delete_index_op, = source_oplog.since(oplog.increment_ts(begin_ts))
		print("attempting", delete_index_op)
		oplog.apply(dest, delete_index_op)
		only_index, = dest.index_deletion_test.stuff.list_indexes()
		assert only_index['name'] == '_id_'
